﻿
using MeshVR;
using MVR.FileManagementSecure;
using SimpleJSON;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Text.RegularExpressions;
using UnityEngine;
using UnityEngine.UI;



namespace Vam_preset_maker
{

   

    public class PresetMaker : MVRScript
    {
        UIDynamicButton _processButton;


        UIDynamicToggle SaveAppearance;
        UIDynamicToggle SaveClothing;
        UIDynamicToggle SaveSkin;
        UIDynamicToggle SaveHair;

        JSONStorableBool SaveAppearanceBool = new JSONStorableBool("AppearancePresets", true);
        JSONStorableBool SaveClothingBool = new JSONStorableBool("ClothingPresets", true);
        JSONStorableBool SaveSkinBool = new JSONStorableBool("SkinPresets", true);
        JSONStorableBool SaveHairBool = new JSONStorableBool("HairPresets", true);

        List<JSONStorableBool> _saveOptions;
        List<string> CreatorsList = new List<string>() { "All" };

        DAZCharacterSelector _dazCharacterSelector;

        CamPositions camPositions = new CamPositions();

        bool _waitingOnScreenShot = false;
        bool firstScreenshot = true;
        bool StopRequested = false;


        CacheSystem cacheSystem;

        FilePaths filePaths;

        JSONStorableStringChooser jssc;
        
        InfoBox infoBox;

        public override void Init()
        {
            //create folders
            if (!FileManagerSecure.DirectoryExists("Custom/PluginData/Chokaphi/PresetMaker/"))
            {
                FileManagerSecure.CreateDirectory("Custom/PluginData/Chokaphi/PresetMaker/");
            }

            if (base.containingAtom.mainController == null)
            {
                SuperController.LogError("Please add this plugin to a PERSON atom.");
                return;
            }

            _dazCharacterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();

            camPositions.StoreCamPosition(null, SuperController.singleton.MonitorRig.transform);
            cacheSystem = new CacheSystem(this);

            JSONStorableString uiInfoBox = new JSONStorableString("uiSetupTimer", "");
            infoBox = new InfoBox(uiInfoBox);

            UIDynamicTextField t = CreateTextField(uiInfoBox);
            RectTransform rt =  t.backgroundImage.GetComponent<RectTransform>();
            Vector2 v = rt.sizeDelta;
            rt.sizeDelta = new Vector2(500, v.y);
            rt.transform.localPosition += new Vector3(250, 0, 0);
            UIDynamic spacerRight = CreateSpacer(true);
            spacerRight.height = 215;

            StartCoroutine(ScanSetup());

        }

        void CreateUI()
        {
            SaveAppearance = CreateToggle(SaveAppearanceBool);
            SaveClothing = CreateToggle(SaveClothingBool);
            SaveSkin = CreateToggle(SaveSkinBool);
            SaveHair = CreateToggle(SaveHairBool);
            _saveOptions = new List<JSONStorableBool>() { SaveAppearanceBool, SaveClothingBool, SaveHairBool, SaveSkinBool };

            jssc = new JSONStorableStringChooser("Creators", CreatorsList, "All", "Select Creator to filter by");
            UIDynamicPopup pop = CreateScrollablePopup(jssc, true);

            _processButton = CreateButton("Process Var Folder", true);
            _processButton.button.onClick.AddListener(() =>
            {
                infoBox.Status = "Processing VAR Files";
                _processButton.button.enabled = false;
                StartCoroutine(ProcessLooks("AddonPackages", "*.json"));
            });

            UIDynamicButton test = CreateButton("WIP Feature - Process Looks", true);

            test.button.onClick.AddListener(() =>
            {
                string filePath = FileManagerSecure.NormalizePath("Saves/Person/appearance/Rory.json");
                JSONNode text = LoadJsonString(filePath);

                StartCoroutine(ProcessLooks("Saves/Person/appearance/", "*.json"));
            });

            infoBox.Message = "UI Created. Ready to start Processing";
        }

        bool inScanExistingVapFiles = false;
        bool inStoreAuthors = false;
        IEnumerator ScanSetup()
        {
            StartCoroutine(ScanExistingVapFiles());
            StartCoroutine(StoreAuthors());

            yield return new WaitWhile(() => inScanExistingVapFiles || inStoreAuthors);
            
            CreateUI();
            infoBox.Status = "UI created";

            yield return null;
        }

        IEnumerator ScanExistingVapFiles()
        {
            inScanExistingVapFiles = true;
            int count = 0;
            foreach (string filePath in GetFiles("Custom/Atom/Person/Appearance/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Appearance";
                infoBox.VapCount = count;
            }
      
            foreach (string filePath in GetFiles("Custom/Atom/Person/Clothing/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Clothing";
                infoBox.VapCount = count;
            }

            foreach (string filePath in GetFiles("Custom/Atom/Person/Hair/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Hair";
                infoBox.VapCount = count;
            }

            foreach (string filePath in GetFiles("Custom/Atom/Person/Skin/", "*.vap"))
            {
                string file = FileManagerSecure.ReadAllText(filePath);
                yield return new WaitForEndOfFrame();
                string hash = cacheSystem.ComputeNodeHash(file);
                yield return new WaitForEndOfFrame();
                cacheSystem.InNodeCache(hash, filePath);
                count++;
                infoBox.VapScan = "Scanning Skin";
                infoBox.VapCount = count;
            }
            cacheSystem.SaveCache();
            inScanExistingVapFiles = false;
            yield return null;
        }

        IEnumerator StoreAuthors()
        {
            inStoreAuthors = true;
            int count = 0;
            foreach (string filePath in GetFiles("AddonPackages", "*.var"))
            {
                count++;
                //uiInfoBox.SetVal("Creating the Author filter, UI will be created once finished. \nVar's found " + count + " Authors found = " + CreatorsList.Count);
                infoBox.Status = "Creating the Author filter";
                infoBox.Message = "UI will be created once scans are finished";
                infoBox.VarCount = count;
                infoBox.AuthorCount = CreatorsList.Count;

                yield return new WaitForEndOfFrame();
                string varName = FileManagerSecure.GetFileName(filePath);

                string[] varNames = varName.Split('.');
                if (!CreatorsList.Contains(varNames[0]))
                {
                    CreatorsList.Add(varNames[0]);
                    yield return new WaitForEndOfFrame();
                }
            }

            CreatorsList.Sort();
            infoBox.AuthorCount = CreatorsList.Count;
            inStoreAuthors = false;

            yield return null;
        }

        IEnumerator StoreCameraPostions()
        {
            Text help = SuperController.singleton.screenshotPreview.GetComponentsInChildren<Text>().Where(x => x.name == "HelpText").FirstOrDefault();
            camPositions.OrigHelpText = help.text;
            camPositions.OrigHelpColor = help.color;

            if (SaveAppearanceBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/AppearancePresets", RecordCamPos);
                help.text = "Set Camera you want to use for Appearance Preset";
                help.color = Color.cyan;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveClothingBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/ClothingPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Clothing Preset";
                help.color = Color.yellow;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveSkinBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/SkinPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Skin Preset";
                help.color = Color.red;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            if (SaveHairBool.val)
            {
                _waitingOnScreenShot = true;
                SuperController.singleton.DoSaveScreenshot("Custom/PluginData/Chokaphi/PresetMaker/HairPresets", RecordCamPos);
                help.text = "Set Camera you want to use for Hair Preset";
                help.color = Color.green;

                yield return new WaitWhile(() => _waitingOnScreenShot);
                yield return new WaitForSecondsRealtime(1);
            }

            help.text = camPositions.OrigHelpText;
            help.color = camPositions.OrigHelpColor;
        }

        void RecordCamPos(string presetType)
        {
            presetType = FileManagerSecure.GetFileName(presetType);
            camPositions.StoreCamPosition(presetType, SuperController.singleton.screenshotCamera.transform);

            _waitingOnScreenShot = false;
        }

        string PresetTypePath(string presetType)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    return "Custom/Atom/Person/Appearance/";

                case "ClothingPresets":
                    return "Custom/Atom/Person/Clothing/";

                case "SkinPresets":
                    return "Custom/Atom/Person/Skin/";

                case "HairPresets":
                    return "Custom/Atom/Person/Hair/";
            }

            return null;
        }


        void SavePreset(string presetType = "ClothingPresets")
        {
            //turn off auto expressions,blink, jaw collision
            containingAtom.GetStorableByID("EyelidControl").SetBoolParamValue("blinkEnabled", false);
            containingAtom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", false);
            containingAtom.GetStorableByID("TongueControl").SetBoolParamValue("tongueCollision", false);
            //Turn on Look at Camera


            List<MeshVR.PresetManagerControl> pmcs = containingAtom.presetManagerControls;
            MeshVR.PresetManagerControl presetControl = pmcs.Where(x => x.name == presetType).FirstOrDefault();
            //set preset name
            presetControl.SetStringParamValue("presetName", filePaths.StringParamValue);
            //store preset
            presetControl.CallAction("StorePreset");
            //now call the screenshot with callback
            _waitingOnScreenShot = true;
        }

        void ScreenShotFinished(string text)
        {
            _waitingOnScreenShot = false;
        }

      

        #region ProcessLooks helpers
        JSONNode LoadJsonString(string filePath)
        {
            try
            {
                string jsonString = FileManagerSecure.ReadAllText(filePath);
                //replace all SELF with var path just to load the person node with internal VAR ref
                string varName = FileManagerSecure.NormalizePath(filePath);
                varName = Regex.Replace(varName, "(.*)(:.+)", "$1");
                jsonString = Regex.Replace(jsonString, "(SELF)(:.+)", varName + "$2");

                return JSON.Parse(jsonString);
            }
            catch (Exception e)
            {
                SuperController.LogMessage("Error: Unable to load JSON File at " + filePath);
                SuperController.LogError(e.Message);
                return new JSONNode();
            }
        }

        void LoadPersonNode(JSONNode personNode)
        {
            JSONClass asObject = personNode.AsObject;
            //load person file
            containingAtom.PreRestore(false, true);
            containingAtom.Restore(asObject, false, true, false, null, false, false, true, false);
            containingAtom.LateRestore(asObject, false, true, false, false, true, false);
            containingAtom.PostRestore(false, true);

            if (SuperController.singleton != null && asObject["id"] != null)
            {
                SuperController.singleton.RenameAtom(containingAtom, asObject["id"]);
                containingAtom.ResetPhysics(false, true);
            }
        }
        #endregion

        bool filter(string filePath)
        {
            string creatorName = Regex.Replace(filePath, @".*\\(.*)(\..*\..*\..*:)(.+)", "$1");

            if (jssc.val == "All")
                return false;

            if (creatorName == jssc.val)
                return false;

            return true;
        }

        void StopProcessRequested()
        {
            infoBox.Message ="Processing stopped";
            SuperController.LogMessage("Preset processing has been Stopped");
            _processButton.button.enabled = true;
            StopRequested = false;
            //reset cam to defalut.
            camPositions.LoadCamPosition(null, SuperController.singleton.MonitorRig.transform);
        }
        IEnumerator ProcessLooks(string path, string pattern)
        {
            int fileCount = 0;
            int lookCount = 0;
            foreach (string filePath in GetFiles(path, pattern))
            {
                infoBox.Message = "Scanning for unprocessed files " + filePath;
                yield return null;
                if (StopRequested || Input.GetKey("space"))
                {
                    StopProcessRequested();
                    yield break;
                }

                if (filter(filePath))
                {

                }
                else if (!cacheSystem.FileInCache(filePath))
                {
                    JSONNode jn = LoadJsonString(filePath);

                    if (jn["atoms"] != null)
                    {
                        fileCount++;
                        JSONNode[] personNodes = jn["atoms"].Childs.Where(x => x["type"].Value == "Person").ToArray();

                        int nodeIndex = 0;
                        foreach (JSONNode personNode in personNodes)
                        {
                            infoBox.Message ="Currently processing file:\n" + FileManagerSecure.GetFileName(filePath) + "\n\nPerson Node" + nodeIndex;
                            yield return new WaitForEndOfFrame();
                            //person node has lets us skip loading duplicate figures
                            if (!cacheSystem.PersonNodeInCache(personNode, filePath))
                            {
                                lookCount++;

                                LoadPersonNode(personNode);

                                DAZCharacterRun dazCharacterRun = _dazCharacterSelector.GetComponent<DAZCharacterRun>();
                                DAZCharacter _dazCharacter = _dazCharacterSelector.selectedCharacter;

                                yield return new WaitUntil(() => _dazCharacter.ready);
                                //wait for physics to finish
                                yield return new WaitWhile(() => dazCharacterRun.IsSimulationResetting());
                                //wait for clothing and any texture loads are finished
                                yield return new WaitWhile(() => SuperController.singleton.loadingIcon.gameObject.activeSelf);
                                //wait 2 seconds for physics to settle
                                yield return new WaitForSeconds(2);

                                //ZERO low value morphs micro values in on versions will cause duplicates any value < 0.01
                                try
                                {
                                    if (_dazCharacterSelector.morphsControlFemaleUI.enabled)
                                    {
                                        _dazCharacterSelector.morphsControlFemaleUI.ZeroNearZero();
                                    }
                                    if (_dazCharacterSelector.morphsControlMaleUI.enabled)
                                    {
                                        _dazCharacterSelector.morphsControlMaleUI.ZeroNearZero();
                                    }
                                }
                                catch
                                {
                                }

                                //for the first screenshot of session allow user to position camera for screenshots
                                if (firstScreenshot)
                                {
                                    firstScreenshot = false;
                                    //record cam position vs head position
                                    Transform headT = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault().transform;
                                    camPositions.SetStartPosition(headT);

                                    yield return StartCoroutine(StoreCameraPostions());
                                }

                                string varName = FileManagerSecure.NormalizePath(filePath);
                                varName = Regex.Replace(varName, "(.*)(:.+)", "$1");

                                filePaths = new FilePaths();
                                filePaths.CreateFileName(varName, personNode["id"], _dazCharacter.isMale);
                                
                                //now store screenshots based on selections
                                foreach (JSONStorableBool jsonBool in _saveOptions)
                                {
                                    //if we have a match in cache skip
                                    if (jsonBool.val) // && !cacheSystem.StorableInCache(filePath, personNode, personNodes.Length, jsonBool.name, value))
                                    {
                                        SavePreset(jsonBool.name);
                                        //store a hash of the saved VAP file. Our storables routine doesnt return the same order as VAM's

                                        if (!cacheSystem.StorableInCache(filePath, personNode, jsonBool.name, filePaths.GetPresetPath(jsonBool.name)))
                                        {
                                            //now store the correct CAM Screenshot
                                            SuperController.singleton.DoSaveScreenshot(filePaths.GetScreenShotPath(jsonBool.name), ScreenShotFinished);

                                            //adjust cam position based on Character height
                                            Transform headT = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault().transform;
                                            camPositions.UpdateHeadDeltas(headT);
                                            camPositions.LoadCamPosition(jsonBool.name, SuperController.singleton.MonitorRig.transform);


                                            //take two frames for camera to reposition.
                                            yield return new WaitForEndOfFrame();
                                            yield return new WaitForEndOfFrame();
                                            //yield return new WaitUntil(() => Input.GetKey(KeyCode.DownArrow));

                                            SuperController.singleton.SetLeftSelect();
                                            yield return new WaitWhile(() => _waitingOnScreenShot);
                                        }
                                        else
                                        {
                                            //if already exists delete created VAP
                                            List<PresetManagerControl> pmcs = containingAtom.presetManagerControls;
                                            PresetManagerControl presetManagerControl = pmcs.Where(x => x.name == jsonBool.name).FirstOrDefault();
                                            PresetManager presetManager = presetManagerControl.GetComponent<PresetManager>();
                                            presetManager.DeletePreset();
                                        }
                                    }

                                }
                                //reset cam for next load. Not needed but makes it look better if watching
                                camPositions.LoadCamPosition(null, SuperController.singleton.MonitorRig.transform);

                                nodeIndex++;
                            }
                        }
                    }
                }

            }

            yield return new WaitForEndOfFrame();
            _processButton.button.enabled = true;
            //reset cam to defalut.
            camPositions.LoadCamPosition(null, SuperController.singleton.MonitorRig.transform);
            //FreeControllerV3 fc = containingAtom.freeControllers.Where(x => x.name == "headControl").FirstOrDefault();
            //SuperController.singleton.FocusOnController(fc);

            infoBox.Message= "Finished Proccessing all VAR json files.\nFiles processed = " + fileCount + "  looks processed = " + lookCount;
        }

        void Update()
        {
            if (!StopRequested && Input.GetKeyDown("space"))
            {
                infoBox.Message = "Preset processing has been requested to stop";
                StopRequested = true;
            }
        }

        static IEnumerable<string> GetFiles(string path, string pattern)
        {
            Queue<string> queue = new Queue<string>();
            queue.Enqueue(path);
            while (queue.Count > 0)
            {
                path = queue.Dequeue();
                try
                {
                    foreach (string subDir in FileManagerSecure.GetDirectories(path))
                    {
                        queue.Enqueue(subDir);
                    }
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                string[] files = null;
                try
                {
                    files = FileManagerSecure.GetFiles(path, pattern);
                }
                catch (Exception ex)
                {
                    SuperController.LogError(ex.Message);
                }
                if (files != null)
                {
                    for (int i = 0; i < files.Length; i++)
                    {
                        yield return files[i];
                    }
                }
            }
        }


        #region


        #endregion

    }

    class FilePaths
    {
        //the name and path to save the preset file. VAM will add 'Preset_' to this for saved VAP file and save in correct folder
        //ex /PresetMaker/MYNAME  becomes  "Custom/Atom/Person/Appearance/PresetMaker/Preset_MYNAME.vap
        public string StringParamValue { get { return presetSubPath + generatedName; } }
        string presetSubPath { get; set; }
        string generatedName { get; set; }
        public string fullVarName { get; private set; }
        public string varName { get; private set; }
        public string authorName { get; private set; }
        

        string appearancePath = "Custom/Atom/Person/Appearance";
        string clothingPath = "Custom/Atom/Person/Clothing";
        string skinPath = "Custom/Atom/Person/Skin";
        string hairPath = "Custom/Atom/Person/Hair";

        public string GetScreenShotPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".jpg";
        }

        public string GetPresetPath(string presetType)
        {
            return PresetTypePath(presetType) + presetSubPath + "Preset_" + generatedName + ".vap";
        }

        string PresetTypePath(string presetType)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    return "Custom/Atom/Person/Appearance";

                case "ClothingPresets":
                    return "Custom/Atom/Person/Clothing";

                case "SkinPresets":
                    return "Custom/Atom/Person/Skin";

                case "HairPresets":
                    return "Custom/Atom/Person/Hair";
            }

            return null;
        }

        public void CreateFileName(string fullVarName, string name = "Person", bool isMale = false)
        {
            string sex = "Female";
            if (isMale)
                sex = "Male";

            this.fullVarName = fullVarName;
            string[] varNames = fullVarName.Split('.');

            authorName = varNames[0];
            varName = varNames[1];

            presetSubPath = string.Concat(new string[]
                {
                    "/",
                    "PresetMaker",

                    "/",
                    authorName,
                    "/",
                });

            if (name.ToLower().Contains("person"))
            {
                name = varName;
            }

            int count = 0;

            string baseName = name + "-" + sex;
            generatedName = baseName;
            string path = presetSubPath + "Preset_" + baseName + ".vap";
            while (FileExists(path))
            {
                count++;
                generatedName = baseName + count;
                path = presetSubPath + "Preset_" + generatedName + ".vap";
            }
        }

        bool FileExists(string path)
        {
            if (FileManagerSecure.FileExists(appearancePath + path) || FileManagerSecure.FileExists(clothingPath + path) || FileManagerSecure.FileExists(skinPath + path) || FileManagerSecure.FileExists(hairPath + path))
            {
                return true;

            }
            return false;
        }
    }

    class CacheSystem
    {
        JSONClass cacheFileJSON = new JSONClass();
        readonly string Cachepath = "Custom/PluginData/Chokaphi/PresetMaker/Cache.json";
        MVRScript mVR;

        public CacheSystem(MVRScript mVR)
        {
            this.mVR = mVR;
            LoadCache();
        }

        public bool FileInCache(string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            if (cacheFileJSON[filePath] == null)
            {
                return false;
            }

           foreach (JSONNode pNode in  cacheFileJSON[filePath].Childs)
            {
                if (pNode.Childs.Count() == 4)
                {
                    //do nothing
                }
                else
                {
                    return false;
                } 
            }

            return true;
        }

        //Used to remove Animations patterns from personNode. Allowing fast hash calculation for duplicate check
        string CleanPersonNodeHash(JSONNode personNode)
        {
            List<JSONNode> animationNodes = new List<JSONNode>();
            foreach (JSONNode storable in personNode["storables"].AsArray)
            {
                if (storable["id"].Value.Contains("Animation"))
                {
                    animationNodes.Add(storable);
                }
            }
            foreach (JSONNode animNode in animationNodes)
            {
                personNode["storables"].AsArray.Remove(animNode);
            }
            return ComputeNodeHash(personNode);
        }

        public bool PersonNodeInCache(JSONNode personNode, string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            string hash = CleanPersonNodeHash(personNode);

            if (cacheFileJSON[filePath][hash] != null)
            {
                if (cacheFileJSON[filePath][hash].Childs.Count() == 4)
                {
                    //SuperController.LogError("PersonNode has 4 entries. Skipping node " + filePath);
                    return true;
                }
            }
           
            return false;
        }

        public bool StorableInCache(string filePath, JSONNode personNode, string presetType, string vapFilePath)
        {
            try
            {
                filePath = FileManagerSecure.NormalizePath(filePath);
                string vapFile = FileManagerSecure.ReadAllText(vapFilePath);

                string cacheSig = ComputeNodeHash(vapFile);
                string personNodeHash = ComputeNodeHash(personNode);

                //if we have a matching personNode and presetType already return true
                if (cacheFileJSON[filePath][personNodeHash][presetType] != null && cacheFileJSON[filePath][personNodeHash][presetType].Value == cacheSig)
                {
                    return true;
                }

                //create cache entry
                bool result = CreateNodeCacheEntry(filePath, personNodeHash, presetType, cacheSig);

                SaveCache();
                return result;
            }
            catch (Exception e)
            {
                SuperController.LogMessage("Error in  StorableInCache routine. " + e.ToString());
                return true;
            }

        }

        public bool InNodeCache(string cacheSig, string filePath)
        {
            filePath = FileManagerSecure.NormalizePath(filePath);
            if (cacheFileJSON["nodeHashes"][cacheSig] != null)
            {
                return true;
            }
            cacheFileJSON["nodeHashes"][cacheSig] = filePath;
            return false;
        }

        bool CreateNodeCacheEntry(string filePath,string personNodeHash, string presetType, string cacheSig)
        {
            if (cacheFileJSON[filePath] == null)
            {
                cacheFileJSON[filePath] = new JSONClass();
            }

            if (cacheFileJSON[filePath][personNodeHash] == null)
            {
                cacheFileJSON[filePath][personNodeHash] = new JSONClass();
            }

            if (cacheFileJSON[filePath][personNodeHash][presetType] == null)
            {
                if (InNodeCache(cacheSig, filePath))
                {
                    cacheFileJSON[filePath][personNodeHash][presetType] = cacheSig;
                    //cacheFileJSON[filePath][personNodeHash][presetType + "Duplicate"].AsBool = true;
                    return true;
                }
                else
                {
                    cacheFileJSON[filePath][personNodeHash][presetType] = cacheSig;
                }
            }

            return false;
        }

        void LoadCache()
        {
            try
            {
                string jsonString = FileManagerSecure.ReadAllText(Cachepath);
                cacheFileJSON = (JSONClass)JSON.Parse(jsonString);
                UpGradeCache();

            }
            catch (Exception e)
            {
                SuperController.LogMessage("Cache not found or was unreadable. New Cache created.");
                cacheFileJSON = new JSONClass();
                UpGradeCache();
            }

            SaveCache();
        }

        public void SaveCache()
        {
            try
            {
                mVR.SaveJSON(cacheFileJSON, Cachepath);
            }
            catch (Exception e)
            {
                SuperController.LogMessage("Failed to save Cache " + e);
            }
        }

        void UpGradeCache()
        {
            if (!cacheFileJSON.HasKey("version") || cacheFileJSON["version"].Value == "1")
            {
                SuperController.LogMessage("New cache created. This will cause duplicate entries for any previously created presets.");
                cacheFileJSON = new JSONClass();
                cacheFileJSON["version"].AsInt = 2;
                cacheFileJSON["nodeHashes"] = new JSONClass();
            }
        }

        #region Hash Functions

        public string ComputeNodeHash(JSONNode node)
        {
            if (node == null)
                return null;

            string hash = ComputeNodeHash(node.ToString());
            return hash;
        }

        public string ComputeNodeHash(string text)
        {
            if (string.IsNullOrEmpty(text))
                return null;

            byte[] raw = System.Text.Encoding.ASCII.GetBytes(text);
            return ComputeNodeHash(raw);
        }

        string ComputeNodeHash(byte[] inputBytes)
        {
            if (inputBytes == null || inputBytes.Length == 0)
                return null;

            using (MD5 md5 = MD5.Create())
            {
                byte[] bytes = md5.ComputeHash(inputBytes);
                StringBuilder builder = new StringBuilder(bytes.Length * 2);
                for (int i = 0; i < bytes.Length; i++)
                {
                    builder.Append(bytes[i].ToString("x2"));
                }
                return builder.ToString();
            }
        }
        #endregion
    }

    class TransformData
    {
        public Vector3 position { get; set; }
        public Quaternion rotation { get; set; }
        public float distance { get; set; }
    }

    class CamPositions
    {
        public string OrigHelpText { get; set; }


        public TransformData VAMCam { get; private set; } = new TransformData();
        public TransformData AppearanceCam { get; private set; } = new TransformData();
        public TransformData ClothingCam { get; private set; } = new TransformData();
        public TransformData SkinCam { get; private set; } = new TransformData();
        public TransformData HairCam { get; private set; } = new TransformData();
        public Color OrigHelpColor { get; set; }
        public TransformData headData { get; private set; } = new TransformData();

        Vector3 deltaP = new Vector3();

        public void UpdateHeadDeltas(Transform transform)
        {
            deltaP = transform.position - headData.position;
        }


        public void SetStartPosition(Transform transform)
        {
            headData.position = transform.position;
            headData.rotation = transform.rotation;
        }


        public void LoadCamPosition(string presetType, Transform target)
        {
            switch (presetType)
            {
                default:
                    target.SetPositionAndRotation(VAMCam.position + deltaP, VAMCam.rotation);
                    return;

                case "AppearancePresets":
                    target.SetPositionAndRotation(AppearanceCam.position + deltaP, AppearanceCam.rotation);
                    return;

                case "ClothingPresets":
                    target.SetPositionAndRotation(ClothingCam.position + deltaP, ClothingCam.rotation);
                    return;

                case "SkinPresets":
                    target.SetPositionAndRotation(SkinCam.position + deltaP, SkinCam.rotation);
                    return;

                case "HairPresets":
                    target.SetPositionAndRotation(HairCam.position + deltaP, HairCam.rotation);
                    return;
            }

        }

        public void StoreCamPosition(string presetType, Transform transform)
        {
            switch (presetType)
            {
                case "AppearancePresets":
                    AppearanceCam.position = transform.position;
                    AppearanceCam.rotation = transform.rotation;
                    AppearanceCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "ClothingPresets":
                    ClothingCam.position = transform.position;
                    ClothingCam.rotation = transform.rotation;
                    ClothingCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "SkinPresets":
                    SkinCam.position = transform.position;
                    SkinCam.rotation = transform.rotation;
                    SkinCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                case "HairPresets":
                    HairCam.position = transform.position;
                    HairCam.rotation = transform.rotation;
                    HairCam.distance = Vector3.Distance(headData.position, transform.position);
                    return;

                default:
                    VAMCam.position = transform.position;
                    VAMCam.rotation = transform.rotation;
                    return;
            }
        }
    }

    public class InfoBox
    {
        JSONStorableString infoBox;
        int varCount = 0;
        int authorCount = 0;
        string status = "Waiting on Scan to Finishe before creating UI";
        string vapScan = "";
        int vapCount = 0;
        string message = "";

        public int VarCount { set { varCount = value; Update();  } }
        public int AuthorCount { set { authorCount = value; Update(); } }
        public string Status { set { status = value; Update(); } }
        public string VapScan { set { vapScan = value; Update(); } }
        public int VapCount { set { vapCount = value; Update(); } }
        public string Message { set { message = value; Update(); } }

        public InfoBox(JSONStorableString uiInfoBox)
        {
            infoBox = uiInfoBox;
        }

        public void Update()
        {
            infoBox.SetVal( String.Format(formated, status, varCount, authorCount, vapScan, vapCount, message));
            
        }

        string formated =
@"{0}                                   {3}
Json's found {1}                                Total existing presets {4}                
Authors found {2}

{5}
        ";


    }


}
